@@ -44,16 +44,19 @@
 #include "event.h"
 #include "adaptors.h"
 #include "object.h"
-#include "common/spi-stateset.h"
+#include "accessible-stateset.h"
 
 #include "accessible-register.h"
 #include "accessible-leasing.h"
 #include "accessible-cache.h"
 
-#include "common/spi-dbus.h"
+#include "spi-dbus.h"
 
 /*---------------------------------------------------------------------------*/
 
+static DBusHandlerResult
+signal_filter (DBusConnection *bus, DBusMessage *message, void *user_data);
+
 SpiBridge *spi_global_app_data = NULL;
 
 /*static Display *bridge_display = NULL;*/
@@ -137,32 +140,34 @@ add_event (const char *bus_name, const char *event)
 static GSList *clients = NULL;
 
 static void
-get_registered_event_listeners (SpiBridge *app)
+tally_event_reply ()
 {
-  DBusMessage *message, *reply;
+  static int replies_received = 0;
+
+  replies_received++;
+  if (replies_received == 3)
+  {
+    if (!clients)
+      spi_atk_deregister_event_listeners ();
+    spi_global_app_data->events_initialized = TRUE;
+  }
+}
+
+static void
+get_events_reply (DBusPendingCall *pending, void *user_data)
+{
+  DBusMessage *reply = dbus_pending_call_steal_reply (pending);
   DBusMessageIter iter, iter_array, iter_struct;
 
-  message = dbus_message_new_method_call (SPI_DBUS_NAME_REGISTRY,
-                                         SPI_DBUS_PATH_REGISTRY,
-                                         SPI_DBUS_INTERFACE_REGISTRY,
-                                         "GetRegisteredEvents");
-  if (!message)
-    return;
-
-  reply = dbus_connection_send_with_reply_and_block (app->bus, message, 5000, NULL);
-  dbus_message_unref (message);
   if (!reply)
-    {
-      spi_global_app_data->events_initialized = TRUE;
-      return;
-    }
+    goto done;
+
   if (strcmp (dbus_message_get_signature (reply), "a(ss)") != 0)
     {
       g_warning ("atk-bridge: GetRegisteredEvents returned message with unknown signature");
-      dbus_message_unref (reply);
-      spi_global_app_data->events_initialized = TRUE;
-      return;
+      goto done;
     }
+
   dbus_message_iter_init (reply, &iter);
   dbus_message_iter_recurse (&iter, &iter_array);
   while (dbus_message_iter_get_arg_type (&iter_array) != DBUS_TYPE_INVALID)
@@ -175,11 +180,104 @@ get_registered_event_listeners (SpiBridge *app)
       add_event (bus_name, event);
       dbus_message_iter_next (&iter_array);
     }
-  dbus_message_unref (reply);
 
-  if (!clients)
-    spi_atk_deregister_event_listeners ();
-  spi_global_app_data->events_initialized = TRUE;
+done:
+  if (reply)
+    dbus_message_unref (reply);
+  if (pending)
+    dbus_pending_call_unref (pending);
+
+  tally_event_reply ();
+}
+
+static void
+get_device_events_reply (DBusPendingCall *pending, void *user_data)
+{
+  DBusMessage *reply = dbus_pending_call_steal_reply (pending);
+  DBusMessageIter iter, iter_array, iter_struct;
+
+  if (!reply)
+    goto done;
+
+  if (strncmp (dbus_message_get_signature (reply), "a(s", 3) != 0)
+    {
+      g_warning ("atk-bridge: get_device_events_reply: unknown signature");
+      goto done;
+    }
+
+  dbus_message_iter_init (reply, &iter);
+  dbus_message_iter_recurse (&iter, &iter_array);
+  while (dbus_message_iter_get_arg_type (&iter_array) != DBUS_TYPE_INVALID)
+    {
+      char *bus_name;
+      dbus_message_iter_recurse (&iter_array, &iter_struct);
+      dbus_message_iter_get_basic (&iter_struct, &bus_name);
+      spi_atk_add_client (bus_name);
+      dbus_message_iter_next (&iter_array);
+    }
+
+done:
+  if (reply)
+    dbus_message_unref (reply);
+  if (pending)
+    dbus_pending_call_unref (pending);
+
+  tally_event_reply ();
+}
+
+static void
+get_registered_event_listeners (SpiBridge *app)
+{
+  DBusMessage *message;
+  DBusPendingCall *pending = NULL;
+
+  message = dbus_message_new_method_call (SPI_DBUS_NAME_REGISTRY,
+                                         ATSPI_DBUS_PATH_REGISTRY,
+                                         ATSPI_DBUS_INTERFACE_REGISTRY,
+                                         "GetRegisteredEvents");
+  if (!message)
+    return;
+
+  dbus_connection_send_with_reply (app->bus, message, &pending, -1);
+  dbus_message_unref (message);
+  if (!pending)
+    {
+      spi_global_app_data->events_initialized = TRUE;
+      return;
+    }
+  dbus_pending_call_set_notify (pending, get_events_reply, NULL, NULL);
+
+  message = dbus_message_new_method_call (SPI_DBUS_NAME_REGISTRY,
+                                         ATSPI_DBUS_PATH_DEC,
+                                         ATSPI_DBUS_INTERFACE_DEC,
+                                         "GetKeystrokeListeners");
+  if (!message)
+    return;
+  pending = NULL;
+  dbus_connection_send_with_reply (app->bus, message, &pending, -1);
+  dbus_message_unref (message);
+  if (!pending)
+    {
+      spi_global_app_data->events_initialized = TRUE;
+      return;
+    }
+  dbus_pending_call_set_notify (pending, get_device_events_reply, NULL, NULL);
+
+  message = dbus_message_new_method_call (SPI_DBUS_NAME_REGISTRY,
+                                         ATSPI_DBUS_PATH_DEC,
+                                         ATSPI_DBUS_INTERFACE_DEC,
+                                         "GetDeviceEventListeners");
+  if (!message)
+    return;
+  pending = NULL;
+  dbus_connection_send_with_reply (app->bus, message, &pending, -1);
+  dbus_message_unref (message);
+  if (!pending)
+    {
+      spi_global_app_data->events_initialized = TRUE;
+      return;
+    }
+  dbus_pending_call_set_notify (pending, get_device_events_reply, NULL, NULL);
 }
 
 static void
@@ -235,12 +333,12 @@ register_application (SpiBridge * app)
 
   /* These will be overridden when we get a reply, but in practice these
      defaults should always be correct */
-  app->desktop_name = SPI_DBUS_NAME_REGISTRY;
-  app->desktop_path = SPI_DBUS_PATH_ROOT;
+  app->desktop_name = ATSPI_DBUS_NAME_REGISTRY;
+  app->desktop_path = ATSPI_DBUS_PATH_ROOT;
 
   message = dbus_message_new_method_call (SPI_DBUS_NAME_REGISTRY,
-                                          SPI_DBUS_PATH_ROOT,
-                                          SPI_DBUS_INTERFACE_SOCKET,
+                                          ATSPI_DBUS_PATH_ROOT,
+                                          ATSPI_DBUS_INTERFACE_SOCKET,
                                           "Embed");
 
   dbus_message_iter_init_append (message, &iter);
@@ -285,8 +383,8 @@ deregister_application (SpiBridge * app)
   dbus_error_init (&error);
 
   message = dbus_message_new_method_call (SPI_DBUS_NAME_REGISTRY,
-                                          SPI_DBUS_PATH_REGISTRY,
-                                          SPI_DBUS_INTERFACE_REGISTRY,
+                                          ATSPI_DBUS_PATH_REGISTRY,
+                                          ATSPI_DBUS_INTERFACE_REGISTRY,
                                           "DeregisterApplication");
   dbus_message_set_no_reply (message, TRUE);
 
@@ -301,28 +399,6 @@ deregister_application (SpiBridge * app)
 
 /*---------------------------------------------------------------------------*/
 
-static void
-exit_func (void)
-{
-  if (!spi_global_app_data)
-    {
-      return;
-    }
-
-  spi_atk_tidy_windows ();
-  spi_atk_deregister_event_listeners ();
-  deregister_application (spi_global_app_data);
-
-  g_free (spi_global_app_data);
-  spi_global_app_data = NULL;
-
-  /* Not currently creating an XDisplay */
-#if 0
-  if (bridge_display)
-    XCloseDisplay (bridge_display);
-#endif
-}
-
 /*---------------------------------------------------------------------------*/
 
 static AtkPlugClass *plug_class;
@@ -353,33 +429,33 @@ socket_ref_state_set (AtkObject *accessible)
   DBusMessageIter iter, iter_array;
   AtkStateSet *set;
 
+  set = atk_state_set_new ();
+
   if (!socket->embedded_plug_id)
-    return NULL;
+    return set;
 
   child_name = g_strdup (socket->embedded_plug_id);
   if (!child_name)
-    return NULL;
+    return set;
   child_path = g_utf8_strchr (child_name + 1, -1, ':');
   if (!child_path)
     {
       g_free (child_name);
-      return NULL;
+      return set;
     }
   *(child_path++) = '\0';
-  message = dbus_message_new_method_call (child_name, child_path, SPI_DBUS_INTERFACE_ACCESSIBLE, "GetState");
+  message = dbus_message_new_method_call (child_name, child_path, ATSPI_DBUS_INTERFACE_ACCESSIBLE, "GetState");
   g_free (child_name);
   reply = dbus_connection_send_with_reply_and_block (spi_global_app_data->bus, message, 1, NULL);
   dbus_message_unref (message);
   if (reply == NULL)
-    return NULL;
+    return set;
   if (strcmp (dbus_message_get_signature (reply), "au") != 0)
     {
       dbus_message_unref (reply);
-      return NULL;
+      return set;
     }
-  set = atk_state_set_new ();
-  if (!set)
-    return  NULL;
+
   dbus_message_iter_init (reply, &iter);
   dbus_message_iter_recurse (&iter, &iter_array);
   do
@@ -421,7 +497,7 @@ socket_embed_hook (AtkSocket * socket, gchar * plug_id)
     {
       DBusMessage *message;
       *(plug_path++) = '\0';
-      message = dbus_message_new_method_call (plug_name, plug_path, SPI_DBUS_INTERFACE_SOCKET, "Embedded");
+      message = dbus_message_new_method_call (plug_name, plug_path, ATSPI_DBUS_INTERFACE_SOCKET, "Embedded");
       dbus_message_append_args (message, DBUS_TYPE_STRING, &path, DBUS_TYPE_INVALID);
       dbus_connection_send (spi_global_app_data->bus, message, NULL);
     }
@@ -478,7 +554,7 @@ get_ancestral_uid (uint pid)
 }
 
 static dbus_bool_t
-user_check (DBusConnection *bus, unsigned long uid)
+user_check (DBusConnection *bus, unsigned long uid, void *data)
 {
   if (uid == getuid () || uid == geteuid ())
     return TRUE;
@@ -498,9 +574,7 @@ new_connection_cb (DBusServer *server, DBusConnection *con, void *data)
   droute_intercept_dbus (con);
   droute_context_register (spi_global_app_data->droute, con);
 
-  new_list = g_list_append (spi_global_app_data->direct_connections, con);
-  if (new_list)
-    spi_global_app_data->direct_connections = new_list;
+  spi_global_app_data->direct_connections = g_list_append (spi_global_app_data->direct_connections, con);
 }
 
 static int
@@ -618,6 +692,25 @@ handle_event_listener_deregistered (DBusConnection *bus, DBusMessage *message,
   remove_events (sender, name);
 }
 
+static void
+handle_device_listener_registered (DBusConnection *bus, DBusMessage *message,
+                                    void *user_data)
+{
+  char *sender;
+  DBusMessageIter iter, iter_struct;
+
+  if (strncmp (dbus_message_get_signature (message), "(s", 2) != 0)
+    {
+      g_warning ("atk-bridge: handle_device_listener_register: unknown signature");
+      return;
+    }
+
+  dbus_message_iter_init (message, &iter);
+  dbus_message_iter_recurse (&iter, &iter_struct);
+  dbus_message_iter_get_basic (&iter_struct, &sender);
+  spi_atk_add_client (sender);
+}
+
 static DBusHandlerResult
 signal_filter (DBusConnection *bus, DBusMessage *message, void *user_data)
 {
@@ -628,7 +721,7 @@ signal_filter (DBusConnection *bus, DBusMessage *message, void *user_data)
   if (dbus_message_get_type (message) != DBUS_MESSAGE_TYPE_SIGNAL)
     return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 
-  if (!strcmp (interface, SPI_DBUS_INTERFACE_REGISTRY))
+  if (!strcmp (interface, ATSPI_DBUS_INTERFACE_REGISTRY))
     {
       result = DBUS_HANDLER_RESULT_HANDLED;
       if (!strcmp (member, "EventListenerRegistered"))
@@ -638,6 +731,16 @@ signal_filter (DBusConnection *bus, DBusMessage *message, void *user_data)
       else
         result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
     }
+  else if (!strcmp (interface, ATSPI_DBUS_INTERFACE_DEVICE_EVENT_LISTENER))
+    {
+      result = DBUS_HANDLER_RESULT_HANDLED;
+      if (!strcmp (member, "KeystrokeListenerRegistered"))
+        handle_device_listener_registered (bus, message, user_data);
+      else if (!strcmp (member, "DeviceListenerRegistered"))
+        handle_device_listener_registered (bus, message, user_data);
+      else
+        result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+    }
 
   if (!g_strcmp0(interface, DBUS_INTERFACE_DBUS) &&
       !g_strcmp0(member, "NameOwnerChanged"))
@@ -796,6 +899,7 @@ adaptor_init (gint * argc, gchar ** argv[])
 
   /* Set up filter and match rules to catch signals */
   dbus_bus_add_match (spi_global_app_data->bus, "type='signal', interface='org.a11y.atspi.Registry', sender='org.a11y.atspi.Registry'", NULL);
+  dbus_bus_add_match (spi_global_app_data->bus, "type='signal', interface='org.a11y.atspi.DeviceEventListener', sender='org.a11y.atspi.Registry'", NULL);
   dbus_connection_add_filter (spi_global_app_data->bus, signal_filter, NULL,
                               NULL);
 
@@ -851,8 +955,54 @@ gnome_accessibility_module_init (void)
 void
 gnome_accessibility_module_shutdown (void)
 {
+  GList *l;
+  GSList *ls;
+
+  if (!spi_global_app_data)
+      return;
+
+  spi_atk_tidy_windows ();
   spi_atk_deregister_event_listeners ();
-  exit_func ();
+
+  deregister_application (spi_global_app_data);
+
+  if (spi_global_app_data->bus)
+    {
+      dbus_connection_remove_filter (spi_global_app_data->bus, signal_filter, NULL);
+      droute_context_unregister (spi_global_app_data->droute, spi_global_app_data->bus);
+      dbus_connection_unref (spi_global_app_data->bus);
+    }
+
+  for (l = spi_global_app_data->direct_connections; l; l = l->next)
+    {
+      droute_context_unregister (spi_global_app_data->droute, l->data);
+      droute_unintercept_dbus (l->data);
+      dbus_connection_unref (l->data);
+    }
+  g_list_free (spi_global_app_data->direct_connections);
+
+  for (ls = clients; ls; ls = ls->next)
+     g_free (l->data);
+  g_slist_free (clients);
+  clients = NULL;
+
+  g_object_unref (spi_global_cache);
+  g_object_unref (spi_global_leasing);
+  g_object_unref (spi_global_register);
+
+  if (spi_global_app_data->main_context)
+    g_main_context_unref (spi_global_app_data->main_context);
+
+  droute_free (spi_global_app_data->droute);
+
+  g_free (spi_global_app_data);
+  spi_global_app_data = NULL;
+
+  /* Not currently creating an XDisplay */
+#if 0
+  if (bridge_display)
+    XCloseDisplay (bridge_display);
+#endif
 }
 
 static gchar *name_match_tmpl =
